from xpanse.base_api import ExApi
from xpanse.api.behavior.v1.risky_flows import RiskyFlowsEndpoint
from xpanse.api.behavior.v1.risk_rules import RiskRulesEndpoint


class RiskyFlowsApi(ExApi, RiskyFlowsEndpoint):
    def __init__(self, session):
        super().__init__(session)

    @property
    def current_version(self):
        return "v1"

    @property
    def v1(self):
        return RiskyFlowsEndpoint(self._api)


class RiskRulesApi(ExApi, RiskRulesEndpoint):
    def __init__(self, session):
        super().__init__(session)

    @property
    def current_version(self):
        return "v1"

    @property
    def v1(self):
        return RiskRulesEndpoint(self._api)


class BehaviorApi(ExApi):
    @property
    def risky_flows(self):
        return RiskyFlowsApi(self._api)

    @property
    def risk_rules(self):
        return RiskRulesApi(self._api)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------
import xpanse

project = "Xpanse"
copyright = "2020, " + xpanse.__author__
author = xpanse.__author__

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.


# The short X.Y version.
version = xpanse.__version__
# The full version, including alpha/beta/rc tags.
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    "rst2pdf.pdfbuilder",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

add_module_names = False
napoleon_google_docstring = True
napoleon_numpy_docstring = False

# -- Options for HTML output -------------------------------------------------

html_theme_options = {
    "description": "The Python interface to the Xpanse Expander API",
    "fixed_sidebar": False,
    "logo": "xpanse_banner.png",
    "touch_icon": "xpanse_banner.png",
    "logo_name": "Xpanse",
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}
html_sidebars = {"**": ["about.html", "navigation.html", "searchbox.html",]}


# PDF Setup
# pdf_documents = [
#     ("index", "XPANSE_SDK_Docs", "Xpanse", xpanse.__author__),
# ]
# pdf_break_level = 0



from typing import Any, Dict

from xpanse.const import V2_PREFIX
from xpanse.endpoint import ExEndpoint
from xpanse.iterator import ExResultIterator


class AnnotationsEndpoint(ExEndpoint):
    """
    Part of the Assets v2 API for handling Annotations.
    See: https://api.expander.expanse.co/api/v1/docs/
    """

    def list_poc(self, **kwargs: Any) -> ExResultIterator:
        """
        Returns a list of points-of-contact.

        Args:
            limit (int, optional):
                Returns at most this many results in a single api call (default: 100, max: 10,000).
            offset (int, optional):
                Returns results starting at this offset (default: 0).
            sort (str, optional):
                Comma-separated string; orders results by the given fields.
                If the field name is prefixed by a -, then the ordering will be descending for that field.
                Allowed values are `id`, `-id`, `created`, `-created`.
            is_assigned (boolean, optional):
                Boolean flag filter to return objects that are assigned to an IP range

        Returns:
            :obj:`ExResultIterator`:
                An iterator containing all of the point-of-contact annotation results. Results can be iterated
                or called by page using `<iterator>.next()`.

        Examples:
            >>> # Return all poc annotations and print each range:
            >>> for res in client.assets.annotations.v2.list_poc():
            ...     for poc in res:
            ...         print(poc)
        """
        return ExResultIterator(
            self._api, f"{V2_PREFIX}/annotation/point-of-contact", kwargs
        )

    def create_poc(self, email: str, **kwargs: Any) -> Dict[str, str]:
        """
        Create a new point-of-contact.

        Args:
            email (str):
                Email address for new point-of-contact.
            firstName (str, optional):
                First name for new point-of-contact.
            lastName (str, optional):
                Last name for new point-of-contact.
            phone (str, optional):
                Phone for new point-of-contact.
            role (str, optional):
                Role for new point-of-contact.

        Returns:
            :obj:`dict`:
                A dictionary containing all of the details about the newly created point-of-contact.

        Examples:
            >>> # Create a new point-of-contact
            >>> poc = client.assets.annotations.v2.create_poc("admin@expanse.co", firstName="Luke", lastName="Skywalker", phone="123-456-7890", role="sysadmin")
        """
        payload = {"email": email}

        if "firstName" in kwargs:
            payload["firstName"] = kwargs["firstName"]
        if "lastName" in kwargs:
            payload["lastName"] = kwargs["lastName"]
        if "phone" in kwargs:
            payload["phone"] = kwargs["phone"]
        if "role" in kwargs:
            payload["role"] = kwargs["role"]
        return self._api.post(
            f"{V2_PREFIX}/annotation/point-of-contact", json=payload
        ).json()

    def delete_poc(self, id: str) -> bool:
        """
        Delete the given point-of-contact, and all connections to other data.

        Args:
            id (str):
                ID for the point-of-contact. Should be a UUID.

        Returns:
            :obj:`boolean`:
                `True` if the point-of-contact was successfully deleted, otherwise `False`.

        Examples:
            >>> # Deletes a point-of-contact
            >>> client.assets.annotations.v2.delete_poc(<id>)
        """
        return (
            True
            if self._api.delete(
                f"{V2_PREFIX}/annotation/point-of-contact/{id}"
            ).status_code
            == 204
            else False
        )

    def get_poc(self, id: str) -> Dict[str, str]:
        """
        Returns the details for the given point-of-contact.

        Args:
            id (str):
                ID for the point-of-contact. Should be a UUID.

        Returns:
            :obj:`dict`:
                A dictionary containing all of the details about a point-of-contact.

        Examples:
            >>> # Return a point-of-contact
            >>> my_range = client.assets.annotations.v2.get_poc(<id>)
        """
        return self._api.get(
            f"{V2_PREFIX}/annotation/point-of-contact/{id}", params={}
        ).json()

    def list_tag(self, **kwargs: Any) -> ExResultIterator:
        """
        Returns a list of tags. Arguments should be passed as keyword args using
        the names below.

        Args:
            limit (int, optional):
                Returns at most this many results in a single api call (default: 100, max: 10,000).
            offset (int, optional):
                Returns results starting at this offset (default: 0).
            sort (str, optional):
                Comma-separated string; orders results by the given fields.
                If the field name is prefixed by a -, then the ordering will be descending for that field.
                Allowed values are `id`, `-id`, `created`, `-created`.
            is_assigned (boolean, optional):
                Boolean flag filter to return objects that are assigned to an IP range
            business_units (str, optional):
                Comma-separated string; Returns only results whose Business Unit's ID falls in the provided list.
                NOTE: If omitted, API will return results for all Business Units the user has permissions to view.
                Also, cannot be used with the business-unit-names parameter.
            business_unit_names (str, optional):
                Comma-separated string; Returns only results whose Business Unit's name falls in the provided list.
                NOTE: If omitted, API will return results for all Business Units the user has permissions to view.
                Also, cannot be used with the business-units parameter.

        Returns:
            :obj:`ExResultIterator`:
                An iterator containing all of the tag annotation results. Results can be iterated
                or called by page using `<iterator>.next()`.

        Examples:
            >>> # Return all tag annotations and print each range:
            >>> for res in client.assets.annotations.v2.list_poc():
            ...     for tag in res:
            ...         print(tag)
        """
        return ExResultIterator(self._api, f"{V2_PREFIX}/annotation/tag", kwargs)

    def create_tag(self, name: str) -> Dict[str, str]:
        """
        Create a new tag.

        Args:
            name (str):
                Name of the tag.

        Returns:
            :obj:`dict`:
                A dictionary containing all of the details about the newly created tag.

        Examples:
            >>> # Create a new tag
            >>> tag = client.assets.annotations.v2.create_tag(name="Wrong registration")
        """
        return self._api.post(f"{V2_PREFIX}/annotation/tag", json={"name": name}).json()

    def delete_tag(self, id: str) -> bool:
        """
        Delete the given tag, and all connections to other data.

        Args:
            id (str):
                ID for the tag. Should be a UUID.

        Returns:
            :obj:`boolean`:
                `True` if the tag was successfully deleted, otherwise `False`.

        Examples:
            >>> # Deletes a taf
            >>> client.assets.annotations.v2.delete_tag(<id>)
        """
        return (
            True
            if self._api.delete(f"{V2_PREFIX}/annotation/tag/{id}").status_code == 204
            else False
        )

    def get_tag(self, id: str) -> Dict[str, str]:
        """
        Returns the details for the given tag.

        Args:
            id (str):
                ID for the tag. Should be a UUID.

        Returns:
            :obj:`dict`:
                A dictionary containing all of the details about a tag.

        Examples:
            >>> # Return a tag
            >>> my_range = client.assets.annotations.v2.get_tag(<id>)
        """
        return self._api.get(f"{V2_PREFIX}/annotation/tag/{id}", params={}).json()



from typing import Any, Dict


def normalize_param_names(kwargs: Dict[str, Any]):
    """
    Replace underscores in kwarg names with hyphens.
    """
    normalized_args = {}
    if "params" in kwargs:
        for k, v in kwargs.get("params", {}).items():
            normalized_args[k.replace("_neq", "!").replace("_", "-")] = v
        kwargs["params"] = normalized_args
    return kwargs



from csv import DictWriter
from json import dump

import click
from xpanse.client import ExClient


def flatten(obj):
    """
    Flattens a json object into a single level using dot notation.
    """
    out = {}

    def _flatten(x, name=''):
        if type(x) is dict:
            for a in x:
                _flatten(x[a], f"{name}{a}.")
        elif type(x) is list:
            for i, a in enumerate(x):
                _flatten(a, f"{name}{i}.")
        else:
            out[name[:-1]] = x

    _flatten(obj)
    return out


@click.command()
@click.argument('file_name', type=click.File('w'))
@click.option('--tags', 'tags', default=None,
    help='Comma-separated string; Returns only results that are associated with the provided tags.')
@click.option('--ip', 'ip', default=None,
    help='Returns the IP  Ranges that match the specified internal CIDR/IP Range/Address.')
@click.option('--include-attribution', 'include_attribution', is_flag=True,
    help='Whether or not attribution details should be included with results.')
@click.option('--include-location', 'include_location', is_flag=True,
    help='Whether or not location details should be included with results.')
@click.option('--include-annotation', 'include_annotation', is_flag=True,
    help='Whether or not annotation details should be included with results.')
@click.option('--include-registration', 'include_registration', is_flag=True,
    help='Whether or not registration details should be included with results.')
@click.option('--include-severity', 'include_severity', is_flag=True,
    help='Whether or not severity details should be included with results.')
def cli(file_name, tags, ip, include_attribution, include_location, include_annotation, include_registration, include_severity):

    # Initialize the Xpanse Client
    client = ExClient()

    # Prepare arguments
    include = []
    if include_attribution:
        include.append("attributionReasons")
    if include_location:
        include.append("locationInformation")
    if include_annotation:
        include.append("annotations")
    if include_registration:
        include.append("relatedRegistrationInformation")
    if include_severity:
        include.append("severityCounts")
    include = ','.join(include) or None

    # Filter and dump all IP Ranges into a list
    ranges = client.assets.ip_range.list(inet=ip,
                                         tag_names=tags,
                                         include=include).dump()

    # Determine the correct output type
    file_type = file_name.name.split(".")[-1]
    if file_type == "json":
        # Dump to json with pretty printing
        dump(ranges, file_name, indent=4, sort_keys=True)
    elif file_type == "csv":
        # Extract the header fields by looking at all of the possible keys
        # Some objects may contain more values than others
        seen_keys = []
        for ip_range in ranges:
            for key in list(flatten(ip_range).keys()):
                if key not in seen_keys:
                    seen_keys.append(key)
        seen_keys.sort()
        # Declare our CSV DictWriter instance
        writer = DictWriter(file_name, fieldnames=seen_keys)

        # Writer row headers to file
        writer.writeheader()

        # For each range, write it to the CSV file.
        for ip_range in ranges:
            writer.writerow(flatten(ip_range))
    else:
        print("Unexpected file type, please use csv or json.")


if __name__ == "__main__":
    cli()



from typing import Any

from xpanse.const import V1_PREFIX
from xpanse.endpoint import ExEndpoint
from xpanse.iterator import ExResultIterator


class UpdatesEndpoint(ExEndpoint):
    """
    Part of the Issues V1 API.
    See: https://api.expander.expanse.co/api/v1/docs/
    """

    def list(self, **kwargs: Any) -> ExResultIterator:
        """
        This endpoint will return a paginated list of issues updates. An issue update is published each time a modification is made to an issue.
        These modifications could include any of the following categories: ProgressStatus, ActivityStatus, Priority, Assignee, and Comment.

        Args:
            limit (int, optional):
                Returns at most this many results in a single api call.
                Default is 100, max is 10000.
            pageToken (str, optional):
                Page token for pagination.
            include (str, optional):
                Comma-separated string; Include the provided fields as part of the serialized result.
                `issue` is the only allowed value and can be provided to include the issue object in the update response.
            createdAfter (str, optional):
                Returns only results created after the provided timestamp (YYYY-MM-DDTHH:MM:SSZ).
            createdBefore (str, optional):
                Returns only results created before the provided timestamp (YYYY-MM-DDTHH:MM:SSZ).
            sort (str, optional):
                Sort by specified properties. Allowed values are `created`, `-created`

        Returns:
            :obj:`ExResultIterator`:
                An iterator containing all of the issue update results. Results can be iterated
                or called by page using `<iterator>.next()`.

        Examples:
            >>> # Return all issues updates between two dates:
            >>> for res in client.issues.updates.v1.list(createdAfter="2020-07-20T00:00:00Z", createdBefore="2020-07-21T00:00:00Z")
            ...     for issue_update in res:
            ...         print(issue_update)
        """
        return ExResultIterator(self._api, f"{V1_PREFIX}/issues/updates", kwargs)



from typing import Any, Dict

from xpanse.const import V2_PREFIX
from xpanse.endpoint import ExEndpoint
from xpanse.iterator import ExResultIterator


class AccountIntegrationsEndpoint(ExEndpoint):
    """
    Part of the Assets v2 API for handling Asset Account Integrations.
    See: https://api.expander.expanse.co/api/v1/docs/
    """

    def list(self, **kwargs: Any) -> ExResultIterator:
        """
        Returns the list of Account Integrations. Arguments should be passed as keyword args using
        the names below.

        Args:
            limit (int, optional):
                Returns at most this many results in a single api call.
                Default is 100, max is 10000.
            pageToken (str, optional):
                Returns results starting at this Page Token.
        Returns:
            :obj:`ExResultIterator`:
                An iterator containing all of the account integration results. Results can be iterated
                or called by page using `<iterator>.next()`.
        Examples:
            >>> # Return all account integrations and print each one:
            >>> for integrations in client.assets.account_integrations.v2.list():
            ...     for integration in integrations:
            ...         print(integration)
        """
        return ExResultIterator(
            self._api, f"{V2_PREFIX}/assets/account-integrations", kwargs
        )



from xpanse.base_api import ExApi
from xpanse.api.targeted_ips.v1.targeted_ips import TargetedIPsEndpoint


class TargetedIPsInnerApi(ExApi, TargetedIPsEndpoint):
    def __init__(self, session):
        super().__init__(session)

    @property
    def current_version(self):
        return "v1"

    @property
    def v1(self):
        return TargetedIPsEndpoint(self._api)


class TargetedIPsApi(ExApi):
    @property
    def targeted_ips(self):
        return TargetedIPsInnerApi(self._api)



from xpanse.base_api import ExApi
from xpanse.api.annotations.v3.tags import TagsEndpoint


class TagsApi(ExApi, TagsEndpoint):
    def __init__(self, session):
        super().__init__(session)

    @property
    def current_version(self):
        return "v3"

    @property
    def v3(self):
        return TagsEndpoint(self._api)


class AnnotationsApi(ExApi):
    @property
    def tags(self):
        return TagsApi(self._api)



from setuptools import setup, find_packages
import os

from xpanse import __version__, __license__, __author__, __email__

try:
    description = open(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.rst")
    ).read()
except:
    description = "Please refer to https://knowledgebase.expanse.co/expander-apis/"
    print("! could not read README.rst file.")

setup(
    name="xpanse",
    version=__version__,
    description="Python library is an interface to the Xpanse Expander API.",
    author=__author__,
    long_description=description,
    author_email=__email__,
    license=__license__,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="xpanse iom",
    packages=["docs", "examples", *find_packages(exclude=["tests"])],
    install_requires=["requests>=2.25.1", "deprecated>=1.2.0"],
    include_package_data=True,
    python_requires=">=3.6",
)



from sys import exit

import click
from xpanse.client import ExClient

ALLOWED_PRIORITIES = ["Critical", "High", "Medium", "Low"]
ALLOWED_STATUSES = ["New", "Investigating", "InProgress", "AcceptableRisk", "Resolved"]
PROGRESS_STATUS = "ProgressStatus"
PRIORITY = "Priority"

@click.command()
@click.option('--issue-type', 'issue_type', multiple=True, default=[],
    help='The type of issue to be filtered on. (ex. "MySQL Server"')
@click.option('--tags', 'tags', multiple=True, default=[],
    help='Tag names to be filtered on.')
@click.option('--status', 'status', default=None, type=click.Choice(ALLOWED_STATUSES),
    help="New status to set all matching issues to.")
@click.option('--priority', 'priority', default=None, type=click.Choice(ALLOWED_PRIORITIES),
    help="New priority to set all matching issues to.")
def cli(issue_type, tags, status, priority):

    # Ensure that multiple updates are not being attempted
    if status is not None and priority is not None:
        exit("Script only supports a single update during each run. Supply --priority OR --status.")
    if status is None and priority is None:
        exit("Script was not supplied an update command. Supply --priority OR --status.")
    # Initialize the Xpanse Client
    client = ExClient()

    # Map issue types from friendly name to id and join to str
    issue_type = ",".join([_issue_name_to_id(client, it) for it in issue_type])

    # Join tags to str
    tags = ",".join(tags)

    # fetch issues
    issues_to_update = client.issues.issues.list(issueTypeId=issue_type, tagName=tags).dump()
    print(f"Script will attepmt to update {len(issues_to_update)} issue to set {'status' if status is not None else 'priority'} to {status if status is not None else priority}")

    # Create update commands
    commands = []
    for issue in issues_to_update:
        if status is not None:
            commands.append((issue.get("id"), status, PROGRESS_STATUS))
        else:
            commands.append((issue.get("id"), priority, PRIORITY))

    # Run Bulk update job
    update_results = client.issues.issues.bulk_update(updates=commands)

    for i, result in enumerate(update_results.get("data")):
        print(f"Operation {i+1} returned status {result.get('status')}")


def _issue_name_to_id(cli, name):
    try:
        types = cli.issues.issue_types.list().dump()
        type_map = {t.get("name"): t.get("id") for t in types}
        return type_map[name]
    except KeyError as err:
        exit(f"Specified Issue Type '{name}' filter was not found")


if __name__ == "__main__":
    cli()



import logging
import os
import platform
from typing import Any, List, MutableMapping, Optional
import sys

import requests
from requests.exceptions import ConnectionError
from urllib3.exceptions import NewConnectionError

from . import __version__
from xpanse.const import (
    XPANSE_BEARER_TOKEN,
    XPANSE_JWT_TOKEN,
    HTTPVerb,
    ID_TOKEN_URL,
)
from xpanse.error import (
    UnexpectedValueError,
    JWTExpiredError,
    XpanseException,
)
from xpanse.api.annotations import AnnotationsApi
from xpanse.api.assets import AssetsApi
from xpanse.api.behavior import BehaviorApi
from xpanse.api.entities import EntitiesApi
from xpanse.api.issues import IssuesApi
from xpanse.api.services import ServicesApi
from xpanse.api.targeted_ips import TargetedIPsApi
from xpanse.api.connectors import ConnectorsApi
from xpanse.utils import normalize_param_names


class ExClient:
    """
    Interface for Xpanse APIs.

    Args:
        url (str, optional):
            The base URL that the paths will be appended onto.  The default
            is ``https://api.expander.expanse.co``
        bearer_token (str, optional):
            The Bearer token for the Xpanse API, which is used to request emphemeral JWT tokens.
            More details at ``https://knowledgebase.expanse.co/expander-apis/#getting``
        jwt (str, optional):
            The JWT for the Xpanse API. Note JWTs are temporary so they are not intended for use
            with long-running applications.
            More details at ``https://knowledgebase.expanse.co/expander-apis/#getting``
        custom_ua (str, optional):
            A custom string can be provided that will be sent within the user-agent header on all
            requests to Xpanse. Final format will be `Xpanse SDK+<custom_ua>/__version__ ...`
        proxies (dict, optional):
            A dictionary detailing what proxy should be used for what transport protocol.
            This value will be passed to the session object after it has been either attached or
            created. For details on the structure of this dictionary, consult the
            :requests:`proxies <user/advanced/#proxies>` section of the Requests documentation.
        verify (bool, optional):
            Whether or not SSL verification should occur. This is `True` by default. Disabling certificate
            verification is strongly discouraged.
            See: https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings InsecureRequestWarning
    """

    """Xpanse URL"""
    _url: str = "https://api.expander.expanse.co"

    """Bearer Token"""
    _bt: Optional[str] = None

    """JWT"""
    _jwt: Optional[str] = None

    """Proxies"""
    _proxies: Optional[str] = None

    """Verify SSL"""
    _verify = True

    """JWT is valid"""
    _jwt_valid: bool = False

    """Max Retry Count"""
    _max_retries: int = 1

    """Error Codes to Retry"""
    _retry_error_codes: List[int] = [503]

    """Vendor Name for UA"""
    _vendor: str = "Xpanse"

    """Product Name for UA"""
    _product: str = "Xpanse SDK"

    """Xpanse SDK Build # for UA"""
    _build: str = "unknown"

    """Python version for UA"""
    _python_version: str = "unknown"

    """OS for UA"""
    _os: str = "unknown"

    """Active Session"""
    _session: requests.Session = requests.Session()

    """Class Methods"""

    def __init__(
        self,
        url=None,
        bearer_token=None,
        jwt=None,
        custom_ua=None,
        proxies=None,
        verify=True,
    ):

        # Format logger
        self._log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )

        # Get install details
        self._set_os()
        self._set_current_python_version()

        # If a custom UA product is desired, include that
        if custom_ua is not None:
            self._product += f"+{custom_ua}"

        # Configure URL and Auth tokens
        if url is not None:
            self._url = url

        if bearer_token is not None:
            self._bt = bearer_token
        else:
            self._bt = os.environ.get(XPANSE_BEARER_TOKEN, None)

        if proxies is not None:
            self._proxies = proxies

        if isinstance(verify, bool):
            self._verify = verify

        if jwt is not None:
            self._jwt = jwt
            self._jwt_valid = True
        else:
            env_jwt = os.environ.get(XPANSE_JWT_TOKEN, None)
            if env_jwt is not None:
                self._jwt = env_jwt
                self._jwt_valid = True
            else:
                self._refresh_jwt()

        if not self._bt and not self._jwt_valid:
            raise UnexpectedValueError(
                "A valid Xpanse Bearer token or JWT token is required."
            )

        # create Session
        self._create_session()

    def _refresh_jwt(self, is_retry: bool = False):
        """
        Refreshes JWT using Bearer token.
        """
        try:
            if self._bt is not None:
                resp = requests.get(
                    f"{self._url}/{ID_TOKEN_URL}",
                    headers={
                        "Authorization": f"Bearer {self._bt}",
                        "Content-Type": "application/json",
                    },
                    timeout=30,
                    proxies=self._proxies,
                    verify=self._verify,
                ).json()
                if "error" in resp or "token" not in resp:
                    raise UnexpectedValueError(
                        f"API returned an error while refreshing JWT: {resp['error']}"
                    )
                else:
                    self._jwt_valid = True
                    self._jwt = resp["token"]
            else:
                raise UnexpectedValueError(
                    "A valid Xpanse Bearer token or JWT token is required."
                )
        except ConnectionError as request_exception:
            if is_retry:
                raise UnexpectedValueError(
                    "Request returned an exception"
                ) from request_exception
            logging.warn(f"ConnectionError encountered during JWT refresh, will retry.")
            self._refresh_jwt(is_retry=True)
        except requests.RequestException as request_exception:
            raise UnexpectedValueError(
                "Request returned an exception"
            ) from request_exception

    def _create_session(self, session: Optional[Any] = None):
        """
        Add JWT auth to session.
        """
        if session is not None:
            self._session = session
        self._session = requests.Session()
        self._session.headers.update(
            {
                "User-Agent": self._generate_user_agent(),
                "Content-Type": "application/json",
                "Authorization": f"JWT {self._jwt}",
            }
        )

        if self._proxies is not None:
            self._session.proxies.update(self._proxies)  # type: ignore

        self._session.verify = self._verify

    def _refresh_session(self):
        """
        Refresh JWT auth if the old token expires.
        """
        if not self._bt:
            raise UnexpectedValueError("No valid Xpanse Bearer token was found.")
        self._refresh_jwt()
        self._session.headers.update({"Authorization": f"JWT {self._jwt}"})

    def _check_response_for_invalid_session(self, resp: requests.Response) -> bool:
        """
        Check response for expired JWT.
        """
        if resp.status_code == 401:
            try:
                resp_json = resp.json()
                if (
                    "detail" in resp_json
                    and resp_json["detail"] == "Signature has expired."
                ):
                    return True
            except ValueError:
                # Some 401 responses are returned without a body
                pass
        return False

    def _set_current_python_version(self):
        """
        Get current installed python version.
        """
        self._python_version = ".".join([str(i) for i in sys.version_info][0:3])

    def _set_os(self):
        """
        GET OS version.
        """
        self._os = f"{platform.version()}"

    def _generate_user_agent(self):
        """
        Generate UA String.
        """
        return f"{self._product}/{__version__} ({self._os}) Python/{self._python_version} {self._vendor}"

    def _request(
        self, method: str, path: str, **kwargs: Any
    ) -> Optional[requests.Response]:
        """
        Request builder.
        """
        retries = 0
        while retries <= self._max_retries:
            try:
                kwargs = normalize_param_names(kwargs)
                self._log.debug(
                    f"REQUEST TO: {method} {self._url}/{path} WITH PAYLOAD: {kwargs}"
                )
                resp = self._session.request(method, f"{self._url}/{path}", **kwargs)
                if resp.status_code < 400:
                    return resp
                else:
                    if self._check_response_for_invalid_session(resp):
                        # JWT has expired, try to generare a new one if a bearer token exists
                        # and retry the request without incrementing our retry counter.
                        if self._bt is not None:
                            self._refresh_session()
                            continue
                        else:
                            raise JWTExpiredError("JWT has expired", resp)
                    else:
                        if resp.status_code in self._retry_error_codes:
                            retries += 1
                            continue
                        else:
                            self._log.error(f"Error response: {resp.text}")
                            return resp
            except (
                ConnectionError,
                NewConnectionError,
            ) as err:
                self._log.error(err)
                retries += 1
                continue
            except XpanseException as err:
                self._log.error(err)
                break
        return None

    def direct_get(self, url: str, **kwargs: Any) -> Optional[requests.Response]:
        """
        Fetches direct without url formatting.
        """
        resp = self._session.request(HTTPVerb.HTTP_GET.value, url, **kwargs)
        return resp

    def get(self, path: str, **kwargs: Any) -> Optional[requests.Response]:
        """
        Initiates an HTTP GET request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`requests.Response`
        """
        return self._request(HTTPVerb.HTTP_GET.value, path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Optional[requests.Response]:
        """
        Initiates an HTTP POST request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`requests.Response`
        """
        return self._request(HTTPVerb.HTTP_POST.value, path, **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Optional[requests.Response]:
        """
        Initiates an HTTP PATCH request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`requests.Response`
        """
        return self._request(HTTPVerb.HTTP_PATCH.value, path, **kwargs)

    def put(self, path: str, **kwargs: Any) -> Optional[requests.Response]:
        """
        Initiates an HTTP PUT request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`requests.Response`
        """
        return self._request(HTTPVerb.HTTP_PUT.value, path, **kwargs)

    def delete(self, path: str) -> Optional[requests.Response]:
        """
        Initiates an HTTP DELETE request using the specified path.  Refer to
        :obj:`requests.request` for more detailed information on what
        keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`requests.Response`

        """
        return self._request(HTTPVerb.HTTP_DELETE.value, path)

    def csv(self, path: str, file_: str, **kwargs: Any) -> bool:
        """
        Initiates an HTTP GET request using the specified path with special handling
        for downloading a csv file. Refer to :obj:`requests.request` for more detailed
        information on what keyword arguments can be passed:

        Args:
            path (str):
                The path to be appended onto the base URL for the request.
            file_ (str):
                The file name with path where the resulting csv file should be saved.
            **kwargs (dict):
                Keyword arguments to be passed to the Requests Sessions request
                method.

        Returns:
            :obj:`boolean`:
                `True` if the download was successful, otherwise `False`.
        """
        try:
            with requests.Session() as s:
                s.headers.update(
                    {
                        "User-Agent": self._generate_user_agent(),
                        "Accept": "text/csv",
                        "Authorization": f"JWT {self._jwt}",
                    }
                )
                resp = s.get(f"{self._url}/{path}", params=kwargs, stream=True)
                if resp.status_code < 400:
                    with open(file_, "wb") as f:
                        for chunk in resp.iter_content(1024 * 1024):
                            f.write(chunk)
                            self._log.debug(f"Writing {path} output to {file_}")
                    return True
                else:
                    self._log.error(f"Error response: {resp.text}")
                    return False
        except requests.exceptions.RequestException:
            self._log.exception(f"Error response")
            return False

    ############################################
    # API Definitions
    ###########################################

    @property
    def annotations(self):
        """Annotations API"""
        return AnnotationsApi(self)

    @property
    def assets(self):
        """Assets API"""
        return AssetsApi(self)

    @property
    def behavior(self):
        """Behavior API"""
        return BehaviorApi(self)

    @property
    def entities(self):
        """Entities API"""
        return EntitiesApi(self)

    @property
    def issues(self):
        """Issues API"""
        return IssuesApi(self)

    @property
    def services(self):
        """Services API"""
        return ServicesApi(self)

    @property
    def targeted_ips(self):
        """Targeted IPs API"""
        return TargetedIPsApi(self)

    @property
    def connectors(self):
        """Connectors API"""
        return ConnectorsApi(self)



from typing import Any

from xpanse.const import V2_PREFIX
from xpanse.endpoint import ExEndpoint
from xpanse.iterator import ExResultIterator


class RegionsEndpoint(ExEndpoint):
    """
    Part of the Assets v2 API for handling Asset Regions.
    See: https://api.expander.expanse.co/api/v1/docs/
    """

    def list(self, **kwargs: Any) -> ExResultIterator:
        """
        Returns the list of Regions. Arguments should be passed as keyword args using
        the names below.

        Args:
            limit (int, optional):
                Returns at most this many results in a single api call.
                Default is 100, max is 10000.
            pageToken (str, optional):
                Returns results starting at this Page Token.

        Returns:
            :obj:`ExResultIterator`:
                An iterator containing all of the region results. Results can be iterated
                or called by page using `<iterator>.next()`.

        Examples:
            >>> # Return all regions and print each one:
            >>> for regions in client.assets.regions.v2.list():
            ...     for region in regions:
            ...         print(region)
        """
        return ExResultIterator(self._api, f"{V2_PREFIX}/assets/regions", kwargs)
